from .session import BaseAsyncSession


class BaseClient:
    DEFAULT_HEADERS = None

    def __init__(self, **session_kwargs):
        self.session = BaseAsyncSession(
            headers=session_kwargs.pop("headers", None) or self.DEFAULT_HEADERS,
            **session_kwargs,
        )

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        self.close()

    def close(self):
        self.session.close()



import time

from playwright.async_api import Error as PlaywrightError, TimeoutError as PlaywrightTimeoutError
from playwright.async_api import async_playwright, Browser, ProxySettings, Page, Playwright
from playwright_stealth import stealth_async
from yarl import URL
from ..base import BasePlaywrightBrowser

from .account import TwitterAccount
from .client import TwitterClient

TWITTER_FUNCAPTCHA_SITEKEY = "0152B4EB-D2DC-460A-89A1-629838B529C9"
TWITTER_FUNCAPTCHA_URL = "https://twitter.com/account/access"


def proxy_url_to_playwright_proxy(proxy: str) -> ProxySettings:
    proxy = URL(proxy)
    return ProxySettings(
        server=f"{proxy.scheme}://{proxy.host}:{proxy.port}",
        password=proxy.password,
        username=proxy.user,
    )


class TwitterUnlocker(BasePlaywrightBrowser):
    # def __init__(self, funcaptcha_solver, **kwargs):
    #     """
    #     :param funcaptcha_solver: Асинхронная функция, решающая фанкапчу. Для ее создания нужно использовать пр
    #     """
    #     self.funcaptcha_solver = funcaptcha_solver
    #     super().__init__(**kwargs)

    async def solve_captcha(self, page: Page, solved_captcha: str) -> bool:
        element = await page.query_selector("#arkose_iframe, input[type='submit'].Button.EdgeButton.EdgeButton--primary")

        if element and element.get_attribute("value") == "Continue to Twitter":
            await element.click()
            print(f'Account successfully unfrozen')

        if element and element.get_attribute('value') == 'Delete':
            await element.click()
            print(f'click delete')

        if element.get_attribute('value') == 'Start':
            await element.click()

            await page.goto('https://twitter.com/account/access')
            await page.wait_for_selector('#arkose_iframe')

        iframe_element = await page.query_selector('#arkose_iframe')
        if not iframe_element:
            if "twitter.com/home" in page.url:
                return True

        iframe = await iframe_element.content_frame()
        script = f'parent.postMessage(JSON.stringify({{eventId:"challenge-complete",payload:{{sessionToken:"{solved_captcha}"}}}}),"*")'
        await iframe.evaluate(script)
        await page.wait_for_load_state(state='networkidle', timeout=10000)


    # TODO должно быть асинхронным контекстным менеджером, который возвращает класс контекста с методом анлокера.
    async def unlock(
            self,
            twitter: TwitterClient,
            *,
            default_timeout: int = 10,
            funcaptcha_solver,
    ):
        # Устанавливаем статус аккаунта, если не установлен прежде
        if twitter.account.status == "UNKNOWN":
            await twitter.establish_status()

        # Если аккаунт не залочен, то ничего не делаем
        if twitter.account.status != "LOCKED":
            return

        # Если залочен, то пытаемся разморозить, решив одну капчи с помощью предоставленного солвера
        proxy = proxy_url_to_playwright_proxy(twitter.session.proxy) if twitter.session.proxy else None
        context = await self._browser.new_context(proxy=proxy)
        cookies = [
            {
                "name": "auth_token",
                "value": twitter.account.auth_token,
                "domain": "twitter.com",
                "path": "/",
            },
        ]
        await context.add_cookies(cookies)
        page = await context.new_page()
        page.set_default_timeout(default_timeout * 1000)
        await stealth_async(page)

        await page.goto(f"https://twitter.com/account/access")
        await page.wait_for_load_state(state='networkidle')

        if "access" not in page.url:
            return

        for _ in range(5):
            await self.solve_captcha(page, solved_funcaptcha)

        # arkose_iframe = await page.query_selector("#arkose_iframe")

        button = await page.wait_for_selector('//*[@id="root"]/div/div[1]/button')
        await button.click()

        ...

        await context.close()

        # Устанавливаем новый статус аккаунта (должен быть GOOD)
        await twitter.establish_status()



from .account import BaseAccount
from .client import BaseClient
from .session import BaseAsyncSession
from .playwright_ import BasePlaywrightBrowser

__all__ = [
    "BaseAccount",
    "BaseClient",
    "BaseAsyncSession",
    "BasePlaywrightBrowser",
]



from pathlib import Path
from typing import Sequence, Iterable

from pydantic import BaseModel

from ..utils import load_lines, write_lines


class BaseAccount(BaseModel):
    auth_token: str | None = None
    username:   str | None = None
    password:   str | None = None
    email:      str | None = None
    name:       str | None = None

    def __init__(
            self,
            auth_token: str,
            *,
            username: str = None,
            password: str = None,
            email: str = None,
    ):
        super().__init__(
            auth_token=auth_token,
            username=username,
            password=password,
            email=email,
        )

    @property
    def short_auth_token(self) -> str:
        start = self.auth_token[:3]
        end = self.auth_token[-3:]
        return f"{start}**{end}"

    @property
    def short_password(self) -> str:
        start = self.password[:2]
        end = self.password[-2:]
        return f"{start}**{end}"

    def __repr__(self):
        return f"{self.__class__.__name__}(auth_token={self.short_auth_token}, username={self.username})"

    def __str__(self):
        return self.short_auth_token

    @classmethod
    def from_file(
            cls,
            filepath: Path | str,
            *,
            separator: str = ":",
            fields: Sequence[str] = ("auth_token", "password", "email", "username"),
    ):
        """
        :param filepath: Путь до файла с данными об аккаунтах.
        :param separator: Разделитель между данными в строке.
        :param fields: Кортеж, содержащий имена полей в порядке их появления в строке.
         Должно содержать как минимум одно поле - auth_token: `("auth_token", )`
        :return: Список аккаунтов.
        """
        accounts = []
        for line in load_lines(filepath):
            data = dict(zip(fields, line.split(separator)))
            data.update({key: None for key in data if not data[key]})
            accounts.append(cls(**data))

        return accounts

    @classmethod
    def to_file(
            cls,
            filepath: Path | str,
            accounts: Iterable["BaseAccount"],
            *,
            separator: str = ":",
            fields: Sequence[str] = ("auth_token", "password", "email", "username"),
    ):
        lines = []
        for account in accounts:
            account_data = []
            for field_name in fields:
                field = getattr(account, field_name)
                field = field if field is not None else ""
                account_data.append(field)
            lines.append(separator.join(account_data))
        write_lines(filepath, lines)



from pydantic import BaseModel


class GoogleAccountData(BaseModel):
    id: int
    name: str
    email: str
    profile_photo_url: str

    def __str__(self):
        return f"{self.email} (id={self.id})"

    @classmethod
    def from_account_list(cls, account_list: list):
        account_data = account_list[1][0]
        return cls(
            name=account_data[2],
            email=account_data[3],
            profile_photo_url=account_data[4],
            id=account_data[10],
        )



from pydantic import BaseModel
from datetime import datetime


class TwitterUserData(BaseModel):
    id: int
    username: str
    name: str
    created_at: datetime
    description: str
    location: str
    followers_count: int
    followings_count: int
    raw_data: dict

    def __str__(self):
        return f"({self.id}) @{self.username}"

    @classmethod
    def from_raw_user_data(cls, data: dict):
        legacy = data["legacy"]
        values = {
            "id": int(data["rest_id"]),
            "username": legacy["screen_name"],
            "name": legacy["name"],
            "created_at": datetime.strptime(legacy["created_at"], '%a %b %d %H:%M:%S +0000 %Y'),
            "description": legacy["description"],
            "location": legacy["location"],
            "followers_count": legacy["followers_count"],
            "followings_count": legacy["friends_count"],
            "raw_data": data,
        }
        return cls(**values)



from curl_cffi import requests

from .account import TwitterAccount

__all__ = [
    "TwitterException",
    "HTTPException",
    "BadRequest",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "RateLimited",
    "TwitterServerError",
    "BadTwitterAccount",
    "BadToken",
    "Locked",
    "Suspended",
]


class TwitterException(Exception):
    pass


class BadTwitterAccount(TwitterException):
    def __init__(
            self,
            account: TwitterAccount,
            custom_exception_message: str = None,
    ):
        self.account = account
        exception_message = f"Bad Twitter account."
        super().__init__(custom_exception_message or exception_message)


class BadToken(BadTwitterAccount):
    def __init__(self, account: TwitterAccount):
        exception_message = f"Bad Twitter account's auth_token."
        super().__init__(account, custom_exception_message=exception_message)


class Locked(BadTwitterAccount):
    def __init__(self, account: TwitterAccount):
        exception_message = f"Twitter account is locked. Captcha required to unlock."
        super().__init__(account, custom_exception_message=exception_message)


class Suspended(BadTwitterAccount):
    def __init__(self, account: TwitterAccount):
        exception_message = f"Twitter account is suspended."
        super().__init__(account, custom_exception_message=exception_message)


class HTTPException(TwitterException):
    """Exception raised when an HTTP request fails.
    """

    def __init__(
            self,
            response: requests.Response,
            response_json: dict = None,
            custom_exception_message: str = None,
    ):
        self.response = response
        self.api_errors: list[dict[str, int | str]] = []
        self.api_codes: list[int] = []
        self.api_messages: list[str] = []

        if response_json is None:
            exception_message = f"{response.status_code}"
            if response.status_code == 429:
                exception_message = (f"{response.status_code} Rate limit exceeded."
                                     f"\nSet TwitterClient(wait_on_rate_limit=True) to ignore this exception.")
            super().__init__(exception_message)
            return

        errors = response_json.get("errors", [])

        if "error" in response_json:
            errors.append(response_json["error"])
        else:
            errors.append(response_json)

        error_text = ""

        for error in errors:
            self.api_errors.append(error)

            if isinstance(error, str):
                self.api_messages.append(error)
                error_text += '\n' + error
                continue

            if "code" in error:
                self.api_codes.append(error["code"])
            if "message" in error:
                self.api_messages.append(error["message"])

            if "code" in error and "message" in error:
                error_text += f"\n{error['code']} - {error['message']}"
            elif "message" in error:
                error_text += '\n' + error["message"]

        if not error_text and "detail" in response_json:
            self.api_messages.append(response_json["detail"])
            error_text = '\n' + response_json["detail"]
        exception_message = f"{response.status_code} {error_text}"
        super().__init__(custom_exception_message or exception_message)


class BadRequest(HTTPException):
    """Exception raised for a 400 HTTP status code.
    """
    pass


class Unauthorized(HTTPException):
    """Exception raised for a 401 HTTP status code.
    """
    pass


class Forbidden(HTTPException):
    """Exception raised for a 403 HTTP status code.
    """
    pass


class NotFound(HTTPException):
    """Exception raised for a 404 HTTP status code.
    """
    pass


class RateLimited(HTTPException):
    """Exception raised for a 429 HTTP status code.
    """
    pass


class TwitterServerError(HTTPException):
    """Exception raised for a 5xx HTTP status code.
    """
    pass



import re
from typing import Literal, Iterable

from yarl import URL
from playwright.async_api import BrowserContext, Request, TimeoutError, Page, Response, Cookie
from playwright_stealth import stealth_async
from .errors import CaptchaRequired, FailedToOAuth2, FailedToLogin, RecoveryRequired
from .account import GoogleAccount
from .utils import check_cookies


PromptType = Literal["consent", "select_account"] | None


def is_valid_google_cookies(cookies: list[dict]) -> bool:
    """
    SID и HSID: Эти cookie содержат цифровые подписи и информацию о последнем входе в систему.
    SSID, APISID, SAPISID: Также содержат информацию об аутентификации и используются в различных сервисах Google для поддержания сессии пользователя.
    __Secure*: Хотя они в первую очередь используются для рекламных целей, они также связаны с твоей учетной записью и могут содержать важную информацию о сессии.
    NID: Используется для хранения настроек пользователя и может содержать информацию, упрощающую доступ к аккаунту.
    GAPS: Этот cookie используется для аутентификации в различных приложениях Google и может содержать важные данные аутентификации.

    Все cookie, содержащие информацию об аутентификации:
        SID
        HSID
        SSID
        APISID
        SAPISID
        OTZ
        NID
        OSID
        LSID
        SIDCC
        ACCOUNT_CHOOSER
        __Secure-1PSIDTS
        __Secure-3PSIDTS
        __Secure-1PSID
        __Secure-3PSID
        __Secure-1PAPISID
        __Secure-3PAPISID
        __Secure-1PSIDCC
        __Secure-3PSIDCC
        __Secure-OSID
        __Host-1PLSID
        __Host-3PLSID
        __Host-GAPS
    """
    return check_cookies(cookies, {"SID", "HSID"})


class GooglePlaywrightBrowserContext:
    # Logining
    _EMAIL_FIELD_XPATH = '//*[@id="identifierId"]'
    _EMAIL_CONFIRMATION_BUTTON_XPATH = '//*[@id="identifierNext"]/div/button'
    _RECAPTCHA_XPATH = '//iframe[@title="reCAPTCHA"]'
    _PASSWORD_FIELD_XPATH = '//*[@id="password"]/div[1]/div/div[1]/input'
    _PASSWORD_CONFIRMATION_BUTTON_XPATH = '//*[@id="passwordNext"]/div/button'
    _RECOVERY_EMAIL_BUTTON_XPATH = '//div[@data-challengeid="5"]'
    _RECOVERY_EMAIL_FIELD_XPATH = '//input[@id="knowledge-preregistered-email-response"]'
    _RECOVERY_EMAIL_CONFIRMATION_BUTTON_XPATH = '//div[@jsname="Njthtb"]/div/button'
    # _NEXT_BUTTON_XPATH = '//button[@jsname="bySMBb"]'
    _PASSKEY_NOT_NOW_BUTTON_XPATH = '//*[@jsname="QkNstf"]/div/div/button'
    _RECOVERY_BUTTON_XPATH = '//div[@id="accountRecoveryButton"]/div/div/a'

    # _PASSKEY_URL = "https://accounts.google.com/signin/v2/passkeyenrollment"
    _PASSKEY_URL_PATTERN = re.compile(r"https://accounts\.google\.com/signin/v2/passkeyenrollment.*")
    _MY_ACCOUNT_URL_PATTERN = re.compile(r"https://myaccount\.google\.com.*")
    _GDS_URL_PATTERN = re.compile(r"https://gds\.google\.com.*")
    _LOGGED_IN_URL_PATTERNS = (_MY_ACCOUNT_URL_PATTERN, _GDS_URL_PATTERN)

    # OAuth2
    _CONTINUE_BUTTON_LOCATOR = '//div[@jsname="uRHG6"]/div/button'
    _ACCOUNT_BUTTON_LOCATOR = '[data-identifier="{email}"]'

    def __init__(
            self,
            context: BrowserContext,
            account: GoogleAccount,
            *,
            stealth: bool = False,
            timeout_to_wait: int = 5_000,
    ):
        self._context = context
        self.account = account
        self.stealth = stealth
        self.timeout_to_wait = timeout_to_wait

        self._logged_in: bool = False
        self._needs_recovery_email: bool = False

    def _account_button_locator(self) -> str:
        return self._ACCOUNT_BUTTON_LOCATOR.format(email=self.account.email.lower())

    async def _new_page(self):
        page = await self._context.new_page()
        if self.stealth: await stealth_async(page)
        return page

    def logged_in(self) -> bool:
        return self._logged_in

    async def login(self):
        if self.account.cookies and is_valid_google_cookies(self.account.cookies):
            await self._context.add_cookies(self.account.cookies)
            self._logged_in = True
            return

        page = await self._new_page()
        try:
            await page.goto("https://accounts.google.com/ServiceLogin")
            await page.locator(self._EMAIL_FIELD_XPATH).type(self.account.email)
            await page.locator(self._EMAIL_CONFIRMATION_BUTTON_XPATH).click()
            await page.wait_for_load_state("networkidle")
            if await page.locator(self._RECAPTCHA_XPATH).count() > 0:
                raise CaptchaRequired("Обнаружена reCAPTCHA.")
            await page.locator(self._PASSWORD_FIELD_XPATH).type(self.account.password)
            await page.locator(self._PASSWORD_CONFIRMATION_BUTTON_XPATH).click()

            # Если в аккаунт с этого IP не входили ранее, то просит ввести recovery_email
            recovery_email_button = page.locator(self._RECOVERY_EMAIL_BUTTON_XPATH)
            try:
                await recovery_email_button.wait_for(timeout=self.timeout_to_wait)
                self._needs_recovery_email = True
            except TimeoutError:
                pass

            if self._needs_recovery_email:
                if not self.account.recovery_email:
                    raise ValueError(f"Needs recovery email")

                await recovery_email_button.click()
                await page.locator(self._RECOVERY_EMAIL_FIELD_XPATH).type(self.account.recovery_email)
                await page.locator(self._RECOVERY_EMAIL_CONFIRMATION_BUTTON_XPATH).click()
                self._needs_recovery_email = False

            await page.wait_for_load_state("load")

            try:
                await page.locator(self._RECOVERY_BUTTON_XPATH).wait_for(timeout=self.timeout_to_wait)
                raise RecoveryRequired("Failed to login Google account."
                                       " Google: We noticed unusual activity in your Google Account."
                                       " To keep your account safe, you were signed out."
                                       " To continue, you’ll need to verify it’s you.")
            except TimeoutError:
                pass

            # Иногда просит установить passkey
            try:
                await page.locator(self._PASSKEY_NOT_NOW_BUTTON_XPATH).click(timeout=self.timeout_to_wait)
            except TimeoutError:
                pass

            await page.wait_for_load_state("load")

            for _ in range(5):
                await page.wait_for_timeout(self.timeout_to_wait / 5)
                if any(url_pattern.search(page.url) for url_pattern in self._LOGGED_IN_URL_PATTERNS):
                    cookies = await self._context.cookies()
                    self._logged_in = is_valid_google_cookies(cookies)
                    break

            await page.close()

            if not self._logged_in:
                raise FailedToLogin("Failed to login Google account: failed to catch auth cookies")

            self.account.cookies = cookies

        except TimeoutError:
            await page.close()
            raise FailedToLogin("Failed to login Google account: unexpected TimeoutError")

    async def oauth2(
            self,
            *,
            client_id: str,
            redirect_uri: str,
            scope: str,
            gsiwebsdk: int = 3,
            access_type: str = "offline",
            response_type: str = "code",
            prompt: PromptType = None,
            include_granted_scopes: bool | str = True,
            enable_granular_consent: bool | str = True,
    ) -> tuple[str | None, str | None]:
        """
        Найдите подобную ссылку: `https://accounts.google.com/o/oauth2/v2/auth?client_id=...`
        Передайте параметры ссылки (после знака вопроса (?)) в метод oauth2
        Метод вернет oauth_code и redirect_url (также содержится в redirect_url)
        :return: oauth_code, redirect_url
        """
        if not self._logged_in:
            await self.login()

        oauth_url = "https://accounts.google.com/o/oauth2/v2/auth"
        params = {
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "scope": scope,
            "gsiwebsdk": gsiwebsdk,
            "access_type": access_type,
            "response_type": response_type,
            "include_granted_scopes": str(include_granted_scopes).lower(),
            "enable_granular_consent": str(enable_granular_consent).lower(),
        }
        if prompt: params["prompt"] = prompt
        oauth_url = str(URL(oauth_url).with_query(params))
        page = await self._new_page()

        oauth_code = None
        redirect_url = None

        async def request_handler(request: Request):
            nonlocal oauth_code
            nonlocal redirect_url

            # Поимка oauth_code и redirect_url основана на знании того, что гугл делает такой редирект:
            # https://developers.google.com/identity/protocols/oauth2/javascript-implicit-flow#redirecting
            if request.url.startswith(redirect_uri):
                redirect_url = URL(request.url)
                oauth_code = redirect_url.query.get(response_type)

        page.on("request", request_handler)

        try:
            await page.goto(oauth_url)
            # TODO Поведение страницы может отличаться, если значение prompt != "consent"
            await page.wait_for_timeout(self.timeout_to_wait)
            await page.locator(self._account_button_locator()).click()
            await page.wait_for_load_state("networkidle")
            try:
                await page.locator(self._CONTINUE_BUTTON_LOCATOR).click(timeout=self.timeout_to_wait)
            except TimeoutError:
                pass
            await page.wait_for_timeout(self.timeout_to_wait)
        except TimeoutError:
            await page.close()
            raise FailedToOAuth2("Failed to OAuth2 Google account: unexpected TimeoutError")

        await page.close()
        return oauth_code, str(redirect_url)



from curl_cffi import requests
from better_proxy import Proxy


class BaseAsyncSession(requests.AsyncSession):
    """
    Базовая асинхронная сессия:
        - Принимает прокси в формате URL и better-proxy.
        - Отключает верификацию SSL сертификатов по умолчанию.
        - По умолчанию устанавливает версию браузера chrome120.
        - По умолчанию устанавливает user-agent под версию браузера chrome120.
    """
    proxy: Proxy | None
    DEFAULT_HEADERS = {
        "accept": "*/*",
        "accept-language": "en-US,en",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36",
        "sec-ch-ua": '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
        "sec-ch-ua-platform": '"Windows"',
        "sec-ch-ua-mobile": "?0",
        "sec-fetch-dest": "empty",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
        "connection": "keep-alive",
        # "accept-encoding": "gzip, deflate, br",
    }
    DEFAULT_IMPERSONATE = requests.BrowserType.chrome120

    def __init__(
            self,
            proxy: str | Proxy = None,
            **session_kwargs,
    ):
        self._proxy = None
        headers = session_kwargs["headers"] = session_kwargs.get("headers") or {}
        headers.update(self.DEFAULT_HEADERS)
        session_kwargs["impersonate"] = session_kwargs.get("impersonate") or self.DEFAULT_IMPERSONATE
        session_kwargs["verify"] = session_kwargs.get("verify", False)
        super().__init__(**session_kwargs)
        self.proxy = proxy

    @property
    def user_agent(self) -> str:
        return self.headers["user-agent"]

    @property
    def proxy(self) -> Proxy | None:
        return self._proxy

    @proxy.setter
    def proxy(self, proxy: str | Proxy | None):
        if not proxy:
            self.proxies = {}
            return

        self._proxy = Proxy.from_str(proxy) if proxy else None
        self.proxies = {"http": self._proxy.as_url, "https": self._proxy.as_url}



"""
Если этот код заработает, фиг я его выложу в паблик...
"""

from typing import Literal
from datetime import datetime

from yarl import URL

from better_automation.base import BaseClient
from better_automation.utils import to_json

from .models import GoogleAccountData

# https://developers.google.com/identity/protocols/oauth2/javascript-implicit-flow#redirecting
PromptType = Literal["consent", "select_account"]


def _referrer_url(**bind_payload) -> str:
    return str(URL("https://accounts.google.com/o/oauth2/v2/auth/oauthchooseaccount").with_query(**bind_payload))


def _continue_url(client_id: str, current_time_with_milliseconds: int) -> str:
    continue_url_params = {
        # TODO параметр as изменчив: as=S-1107092599:1704645143377026
        'as': f'S118374196:{current_time_with_milliseconds}',
        #
        'authuser': 'unknown',
        'client_id': client_id,
        # TODO параметр part изменчив
        'part': 'AJi8hAM10gN03GEzqyv-T5HOnWXaAgNCkRNxL7W3qkj40wVxz_iW3KVaBsD9Ln3Z0pS0C8oYzllffQO-klibpEvBTKubiNDrlHhrSuSRKTmaOanJXoE-CXvVcZtB79DoDltu28ZF-uRn6DuqIUa59YsG5Ucf05X1MQT3WHeiJAy5FVCxOuVuO02I4t54i1rc0XEgFsu5btbHsZbvC4oi95rbMlOK2GcrDS5_puCM7Ye1QVYetODMYfRhFrbBUzxt95nrkrRmBxpWcikoZqktK83ZVvY4IzeqfwcmJXVxfIq5bbauooCFs0SuM-b9rsNNhYbgNtsemfz7KkwAV5xZwsN_eVkGhVLwubSGYm0hRFEStIKyxafrmO2WE49-6-XQUHszlkB2V89wSftaPYhTpSWCu8RGFMg4eUrhDwK4YLvIpLxIbUhwcOhAvEQHPGtVxIIb2P8dktoXbqgpQ0M-4mJj2nUEJegrHg',
        'theme': "glif#",
    }
    return str(URL("https://accounts.google.com/signin/oauth/consent").with_query(**continue_url_params))


def _payload(client_id: str, current_datetime: datetime) -> dict:
    current_time_with_milliseconds = int(current_datetime.timestamp() * 1000000)
    current_time = int(current_datetime.timestamp())
    flow_name = "GeneralOAuthFlow"
    bg_request = [
        "oauth-account-chooser",
        "<m1dqVw8CAAbBwkrJ4cKN8og_h42xmOz0ADQBEArZ1M2W1E1GYLsYZv5-DiRc8bGCRjnJRmHOvDwBLEEYPP_kxVrHjdrCGzlJBYpGZadgzQAAAICdAAAABKcBB1YGMJk3Cy12utdgB5x5dU1KjKSuaHLwISrKnxWwbrY8Yk_yYQMAm2vP8tRHQBvLJHtyGrHaNVsW33HKY-Gh4gjr6jtb521K9HRUinPDWkb_gXSyb5_E-F5DrtHH4HPpEKNX0Fbrljj5STVY3Moq4_OBDS0M82yg0FcCJHGoB5sHUCcxyGdqxeHIQ_hlzJf9YmdNHH2DR58Cuww6ocANknEmV41ixG6_a2D8Y_bZTIb_dzmK3gdGtCmxSJo3UPsVUKzgAcN4JuL7jdbOIYt0ZkdXT_S6Uq1M0iCtvNCW0MnYAAK1mjaoh7s94MhLpVm6igk546q_OhWuLKQ4gk2YW-XzDL8zN22eJSrkGo1tv7q_3G7BEdM3WT840wBtcSmvBXMaBDGhwhqp9e19I4Uld9bHIdJQ_LhFUXSw4ZmKv7e1ifpJnhlZZjK-fqrZ_gouJij-S-S4BRc_ClVL3peUYZ_mZt8CP_vWjXMhvKrAziJc-CSyF4-JtQsixlFvTkp6ec3VsFtLoS4Bbq2K0byQq0CqkZstCXWbECnonWznR_YvIh5x_vDYzQcIWwNVpnuS3f3F9NWlNkzfSa8RW-J2Ze-IPd5eHInhcQ2mJIh_sh05jiZ5Mn4abPEBKpNd2bZb2oV8-8DxJ-77njqzNjHGBaXwsCt5oU_8OGVucSeRssm5hxXYZwGw-zlvIpbG0BSdUR76jisEMlkZM60B6Y-EJyaWuinMkKJZdeaKTMaDxwys0QvajH8u4OwZ-wcgh77aetN_wzi2ERM3-hvQ8H6TvcrLPNp09-TKZQYxkG37CS545mN3TQWO2OU65e-3EMht7dXUBgSRn8BHvJbO1JAadPdhX_yGF_iTM2pAWTTPdLselh0EKZX7YrBxkMTNmoYbgv8vCKyauj-oOW-QUmh80CLzKabfl5JWHJbJIH1mc5S_4eQWU4QV-_9GhiDAXLLS0ccM4m4JYEoypjm4OuDVMzUwCmJLqDJRAE8hJmVRddGL4Xl6sLwPmDGu-5912PTX7O4ObFpxqTs3sjXkXr5oocRBbWDnIMlD7vLEykQRxZbRwKH-lOfxkoz77o8yGvWRJg_tNVW_Io0oUymAY1fgxUrm0jN-T7SFP2r6Aqpf4Y_p-2kkk83Q0lWVpDpyhCfKD_1nWKj9QG5aFoeWm-cUCbU7HsxvHEUjQ5g01Ps8Wl67TBR_gPOECXbfYS0p0yfiF-caZvT1K11w_u42-zwgaIOnkX8R1aEP-xQxtd0OCWyoMB-jtBy4GUQeU7ro6BI-gfOKK1MTfvu8kdt1gvEoUaVTcojT4SK77djjNxq5MXIGMLnfDPnlU0P7ghN1PL_yXFbGsqH-SITLOruq_50io8tfu4XjbKGWBgrHBf-98GR-WUNSeP6_3be3CfG1qKUOmZgqsXWA6OmDO7ugwH8ITs-QC1BRNuxY-fen7G0qR2UTlqy9Mn9dDRxG69vcn11JwyRvY1uXX64qaPu-Z0y5EJzkmjq9sovjf6TY-eAHSfUtx1wvu1d_kkMifwB3SdTx6-8FxaNNBqdIzTmwYBLOW6OjuRq-sdXOsxaL2vNSCQ_dRBHD-FQJDqgAggHGepeu3DIU1ojkPJF40v4kImi7AmvrQaGet_QiU3bBgRXCVjeeYSK_7BZwI24AMGFZfrsnFG2MTcUHcINpTy8zKmcYIzrqVXqgru7Iva4qLFBF4TgWaAxpWf22zsKk94MOY69ikY8AK0KDKLKzq0I3qYavfNk--ERdTD0OnJy7qk86C5rsY9AVVgOiXi6QWoZqPRlukeD49SDh9F6b4UxBWnLgJPQhLjPATe5aqJ1KJ5ouHlOZVq1fD88Q9PJI5YHMIMAToA79_y_2Uj2pB2KzpfFvBoOTKZH_tcKY0Vdi_yw5LyRoZLE8HNgQTaYa642kouRjTrReWqHlHeB9yh8THFygQi83Jraeg_qswUkB4H2srpzE1TAb8nbWT8gdUZoFhqIk33MPWFfOnTiYDBVzEpWafMewf40mroxK_pFmCdjJZezTh1JlUVXWXJqOWHRaAWewSzmfja0jyC7RZTDvmUC8NwVP4yLWDER-v2t5JKkxyR0JL8vSW-TjGKx14SAj_oJsbq2bMg"
    ]
    f_req = [
        # TODO этот параметр изменчив
        "AEThLlwSApN39ij_LIvwDx90E4xGGvn4fbEegbNpEGiMeh2AhS7LOM3TH5ZRCty34T-TkNNRZC5YNBcthqe7wXWzQuQpBVj-OTuGh8FvjTp0s4y1wzS3upCCk3CnWh67bIQ90EjTwMf42ccWzwQiWoxfP_FLYKXZRstfsbzC8UcOKGzSKWrVWlBZf--YyxOy4b0cGvha6lnX-b6yFQ2-HRANvow1ndMPSboPgkTp6wS_9i8e8ayg9K7kVt8ikyIHsrI1T9N4mkNgpH3v4JzoJdbv9HmIXPtI18CahN1K8rme995o9Tbijp_qWsQkwaxn_21em1-sarkFKdOsaQZzhKQ5tAMDbSRl-moN6_tMqqo7lvVw7A6e6C69u-5rn7wdsfI8wiKi3UtuAiZ7foHMK03K49FlOXU4vDAa2gWD-9xitIneRPYqreqNvJe88XckT4mSMlx9sMYpWKfOemVbcmYKV_7NA_vBaXsea8QvOCuXySg-NirAarTZ9kbcnTJ3fapfViyDu8db",
        # TODO это device info
        0, 0, None, [None, None, [2, 1, None, 1,
                                  # TODO собрать ссылку
                                  "https://accounts.google.com/signin/oauth?client_id=968583257586-kljtp79kj8nc53gocd7lmo3sfgbd2i1f.apps.googleusercontent.com&scope=https%3A%2F%2Fwww.googleapis.com%2Fauth%2Fuserinfo.email&redirect_uri=https%3A%2F%2Fairdrop.gomble.io%2Flogin&prompt=consent&gsiwebsdk=3&access_type=offline&response_type=code&include_granted_scopes=true&enable_granular_consent=true",
                                  None, None, 4, None,
                                  flow_name,
                                  None, None, 1], 10, [None, client_id, None,
                                                       "!ChREeHdtVjRYM0I4bkFDMUFBSFJ4YhIfVThjbHR4Q3NHaTBaOERFdWhZOThQYzkxdF9VSHpoZw∙AHkTZLMAAAAAZZsLtG-mLSf6L6c9A0qw-Bu0HbeCQNmD",
                                                       None, None, None, None, None, None, None,
                                                       # TODO вставлять верную ссылку
                                                       "https://airdrop.gomble.io",
                                                       f"S118374196:{current_time_with_milliseconds}",
                                                       0, None, None, None, None, [5,
                                                                                   # TODO another_client_id
                                                                                   "77185425430.apps.googleusercontent.com",
                                                                                   [
                                                                                       "https://www.google.com/accounts/OAuthLogin"],
                                                                                   None, None,
                                                                                   # TODO device_id
                                                                                   "44a22fdf-84ac-4f1f-b2cc-4911e62f955f",
                                                                                   None, None, None, None, None,
                                                                                   None, None, None, None, None,
                                                                                   None, None, None, None, None,
                                                                                   None, 5], None, None, None, 25,
                                                       None, None, None, 1, None, None, None,
                                                       [[None, 202], [None, 9315]], None, 1, None, None, 1], None,
                     None, None, 1, None, None, None, None, None, None, None, None, [], None, None, 3]]
    device_info = [
        None, None, None, None, None,
        "EN", None, None, None,
        flow_name, None,
        [None, client_id, None,
         # TODO этот параметр изменчив
         "!ChREeHdtVjRYM0I4bkFDMUFBSFJ4YhIfVThjbHR4Q3NHaTBaOERFdWhZOThQYzkxdF9VSHpoZw∙AHkTZLMAAAAAZZsLtG-mLSf6L6c9A0qw-Bu0HbeCQNmD",
         None, None, None, None, None, None, None,
         "https://airdrop.gomble.io", f"S118374196:{current_time_with_milliseconds}", 0, None, None, None, None,
         # TODO another_client_id
         [5, "77185425430.apps.googleusercontent.com",
          ["https://www.google.com/accounts/OAuthLogin"], None, None,
          # TODO device_id
          "44a22fdf-84ac-4f1f-b2cc-4911e62f955f",
          None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, 5], None,
         None, None, 25, None, None, None, 1, None, None, None, [[None, 202], [None, 9315]], None, 1, None, None,
         1], None, None, None, None, 1, None, 0, 1, "", None, None, 1, 1
    ]

    payload = {
        # TODO параметры at и azt изменчивы: AFoagUVZK44l9a8PZIB4DhOheiGdHOkfDw:1704645143440
        'at': f'AFoagUVij2Eb9qDG4z_VA5IKP9oH-I2auw:{current_time}',
        'azt': f'AFoagUVij2Eb9qDG4z_VA5IKP9oH-I2auw:{current_time}',
        'checkConnection': 'youtube:332',
        'checkedDomains': 'youtube',
        'cookiesDisabled': 'false',
        'flowName': flow_name,
        'gmscoreversion': 'undefined',
        'continue': _continue_url(client_id, current_time_with_milliseconds),
        'bgRequest': to_json(bg_request),
        'f.req': to_json(f_req),
        'deviceinfo': to_json(device_info),
    }
    return payload


class GoogleOAuthClient(BaseClient):
    DEFAULT_HEADERS = {
        'authority': 'accounts.google.com',
    }

    def __init__(self, cookies: list[dict], **session_kwargs):
        cookies = [(cookie['name'], cookie['value']) for cookie in cookies]
        super().__init__(cookies=cookies, **session_kwargs)
        self.account_data: GoogleAccountData | None = None

    @property
    def _x_chrome_id_consistency_request(self) -> str:
        if self.account_data is None:
            raise ValueError("First request account data")

        x_chrome_id_consistency_request_dict = {
            "version": 1,
            # TODO another_client_id
            "client_id": "77185425430.apps.googleusercontent.com",
            # TODO device_id
            "device_id": "44a22fdf-84ac-4f1f-b2cc-4911e62f955f",
            "sync_account_id": self.account_data.id,
            "signin_mode": "all_accounts",
            "signout_mode": "show_confirmation",
        }

        return ','.join(f"{key}={value}" for key, value in x_chrome_id_consistency_request_dict.items())

    async def request_account_data(self):
        url = "https://accounts.google.com/ListAccounts"
        params = {
            "listPages": 0,
            "pid": 243,
            "gpsia": 1,
            "source": "ogb",
            "atic": 1,
            "mo": 1,
            "mn": 1,
            "hl": "en",
            "ts": 157,
        }
        payload = {}
        headers = {
            'origin': 'https://ogs.google.com',
            'referer': 'https://ogs.google.com/',
            'x-client-data': 'CIq2yQEIpbbJAQipncoBCNX/ygEIk6HLAQib/swBCIWgzQEIoe7NAQiD8M0BGPTJzQEYp+rNAQ=='
        }
        response = await self.session.request("POST", url, headers=headers, data=payload, params=params)
        self.account_data = GoogleAccountData.from_account_list(response.json())
        print(self.account_data)

    async def oauth_2(
            self,
            client_id: str,
            redirect_uri: str,
            scope: str,
            gsiwebsdk: int = 3,
            access_type: str = "offline",
            response_type: str = "code",
            state: str = None,
            prompt: PromptType = None,
            login_hint: str = None,
            include_granted_scopes: bool = None,
            enable_granular_consent: bool = None,
    ):
        if self.account_data is None:
            await self.request_account_data()

        url = "https://accounts.google.com/_/signin/oauth"
        params = {
            "authuser": 0,
            "hl": "en",
            "_reqid": 185158,  # 70399 - плохой
            "rt": "j",
        }

        current_datetime = datetime.now()
        payload = _payload(client_id, current_datetime)
        bind_payload = {
            "client_id": client_id,
            "scope": scope,
            "redirect_uri": redirect_uri,
            "prompt": prompt,
            "gsiwebsdk": gsiwebsdk,
            "access_type": access_type,
            "response_type": response_type,
            "service": "lso",
            "o2v": 2,
            "theme": "glif",
        }
        if state: bind_payload["state"] = state
        if state: bind_payload["prompt"] = prompt
        if login_hint: bind_payload["login_hint"] = login_hint
        if include_granted_scopes: bind_payload["include_granted_scopes"] = str(include_granted_scopes).lower()
        if enable_granular_consent: bind_payload["enable_granular_consent"] = str(enable_granular_consent).lower()
        payload.update(bind_payload)
        headers = {
            'authority': 'accounts.google.com',
            'content-type': 'application/x-www-form-urlencoded;charset=UTF-8',
            'google-accounts-xsrf': '1',
            'origin': 'https://accounts.google.com',
            'referer': _referrer_url(**bind_payload),
            'sec-ch-ua': '"Not_A Brand";v="8", "Chromium";v="120", "Google Chrome";v="120"',
            'sec-ch-ua-arch': '"x86"',
            'sec-ch-ua-bitness': '"64"',
            # 'sec-ch-ua-full-version': '"120.0.6099.199"',
            # 'sec-ch-ua-full-version-list': '"Not_A Brand";v="8.0.0.0", "Chromium";v="120.0.6099.199", "Google Chrome";v="120.0.6099.199"',
            # 'sec-ch-ua-mobile': '?0',
            # 'sec-ch-ua-model': '""',
            # 'sec-ch-ua-platform': '"Windows"',
            # 'sec-ch-ua-platform-version': '"15.0.0"',
            # 'sec-ch-ua-wow64': '?0',
            # 'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.36',
            'x-chrome-id-consistency-request': self._x_chrome_id_consistency_request,
            'x-client-data': 'CIq2yQEIpbbJAQipncoBCNX/ygEIk6HLAQib/swBCIWgzQEI3L3NAQiP4c0BCOLszQEIoe7NAQi87s0BCIrvzQEIg/DNAQiG8M0BCLLxzQEIp/LNARj0yc0BGKfqzQE=',
            'x-same-domain': '1'
        }

        response = await self.session.request("POST", url, headers=headers, data=payload, params=params)

        print(response.text)



from functools import cache
from typing import Literal
import base64
import json


@cache
def _encode_x_properties(properties: str) -> str:
    return base64.b64encode(properties.encode('utf-8')).decode('utf-8')


def encode_x_properties(properties: dict) -> str:
    return _encode_x_properties(json.dumps(properties, separators=(',', ':')))


@cache
def decode_x_properties(encoded_properties: str) -> dict:
    decoded_bytes = base64.b64decode(encoded_properties.encode('utf-8'))
    return json.loads(decoded_bytes.decode('utf-8'))


def create_guild_x_context_properties(
        location_guild_id: int | str,
        location_channel_id: int | str,
        location_channel_type: int = 0,
        location: Literal["Accept Invite Page", "Join Guild"] = "Accept Invite Page",
) -> str:
    x_context_properties = {
        "location_guild_id": str(location_guild_id),
        "location_channel_id": str(location_channel_id),
        "location_channel_type": location_channel_type,
        "location": location,
    }
    return encode_x_properties(x_context_properties)


def create_x_super_properties(
        user_agent: str,
        client_build_number: int,
        *,
        browser: str = "Chrome",
        browser_version: str = "110.0.0.0",
        os: str = "Windows",
        os_version: str = "10",
        system_locale: str = "en-US",
        # system_locale: str = "ru-RU",
) -> str:
    x_super_properties = {
        'browser': browser,
        'browser_user_agent': user_agent,
        'browser_version': browser_version,
        'client_build_number': client_build_number,
        'client_event_source': None,
        'device': '',
        'os': os,
        'os_version': os_version,
        # 'referrer': 'https://discord.com/',
        # 'referring_domain': 'discord.com',
        'referrer': '',
        'referring_domain': '',
        'referrer_current': '',
        'referring_domain_current': '',
        'release_channel': 'stable',
        'system_locale': system_locale}
    return encode_x_properties(x_super_properties)


from datetime import datetime
from typing import Iterable


def check_cookies(
        cookies: list[dict],
        cookies_to_check: Iterable[str],
) -> bool:
    # Создание словаря для хранения найденных cookie
    found_cookies = {name: False for name in cookies_to_check}

    # Текущее время в формате Unix timestamp
    current_timestamp = datetime.now().timestamp()

    for cookie in cookies:
        # Если имя cookie есть в списке для проверки и его срок ещё не истёк
        if cookie['name'] in cookies_to_check:
            if 'expires' not in cookie:
                found_cookies[cookie['name']] = True
            elif current_timestamp < cookie['expires']:
                found_cookies[cookie['name']] = True

    # Проверяем, все ли нужные cookie найдены и действительны
    return all(found_cookies.values())



from .client import DiscordClient
from .account import DiscordAccount, DiscordAccountStatus
from .utils import to_invite_code
from . import errors

__all__ = [
    "DiscordClient",
    "DiscordAccount",
    "DiscordAccountStatus",
    "to_invite_code",
    "errors"
]



class GoogleError(Exception):
    pass


class FailedToLogin(GoogleError):
    """Исключение, вызываемое если не получилось залогиниться по неизвестной причине."""
    pass


class CaptchaRequired(FailedToLogin):
    """Исключение, вызываемое если не получилось залогиниться из-за reCAPTCHA."""
    pass


class RecoveryRequired(FailedToLogin):
    """
    Исключение, вызываемое если не получилось залогиниться из-за того, что требуется восстановление аккаунта.

    Google: We noticed unusual activity in your Google Account.
        To keep your account safe, you were signed out.
        To continue, you’ll need to verify it’s you
    """
    pass


class FailedToOAuth2(GoogleError):
    """Исключение, вызываемое если не получилось авторизоваться через OAuth2 по неизвестной причине."""
    pass



from enum import StrEnum

from pydantic import Field

from ..base import BaseAccount

DISCORD_AUTH_TOKEN_PATTERN = r"^[A-Za-z0-9+._-]{72}$"


class DiscordAccountStatus(StrEnum):
    BAD_TOKEN = "BAD_TOKEN"  # (401, 403?)
    UNKNOWN   = "UNKNOWN"
    GOOD      = "GOOD"

    def __str__(self):
        return self.value


class DiscordAccount(BaseAccount):
    auth_token:     str = Field(default=None, pattern=DISCORD_AUTH_TOKEN_PATTERN)
    status:         DiscordAccountStatus = DiscordAccountStatus.UNKNOWN,
    is_spammer:     bool = False
    is_quarantined: bool = False
    phone:          str | None = None
    # TODO TOTP_secret



from base64 import b64decode
from enum import StrEnum
import json

from pydantic import Field

from ..base import BaseAccount

TWITTER_AUTH_TOKEN_PATTERN = r"^[a-f0-9]{40}$"


class TwitterAccountStatus(StrEnum):
    BAD_TOKEN = "BAD_TOKEN"  # (401) 32
    UNKNOWN = "UNKNOWN"
    SUSPENDED = "SUSPENDED"  # (403) 64, (200) 141
    LOCKED = "LOCKED"  # (403) 326
    GOOD = "GOOD"

    def __str__(self):
        return self.value


class TwitterAccount(BaseAccount):
    auth_token: str = Field(default=None, pattern=TWITTER_AUTH_TOKEN_PATTERN)
    id: int | None = None
    ct0: str | None = None
    status: TwitterAccountStatus = TwitterAccountStatus.UNKNOWN
    # TODO user_data

    @classmethod
    def from_cookies(
            cls,
            cookies: dict | str,
            *,
            base64: bool = False,
            **kwargs,
    ) -> "TwitterAccount":
        if base64:
            cookies = json.loads(b64decode(cookies).decode('utf-8'))
        elif isinstance(cookies, str):
            cookies = json.loads(cookies)

        auth_token = None
        ct0 = None

        for cookie in cookies:
            if cookie['name'] == 'auth_token':
                auth_token = cookie['value']
            elif cookie['name'] == 'ct0':
                ct0 = cookie['value']

        if auth_token is None:
            raise ValueError("auth_token not found in cookies.")

        ct0 = kwargs.pop("ct0", None) or ct0

        account = cls(auth_token, **kwargs)
        account.ct0 = ct0
        return account
